% A Graphical User Interface for Constraint Logic Programming
% http://contraintes.inria.fr/~fages/CLPGUI
% Copyright (C) 2001,2002, 2006 INRIA Francois Fages
% License LGPL-2.1

% Version 2.1.6
% file clpgui.pl Prolog library to use with the graphical user interface program CLPGUI.java

% 27 Nov 2001 Version 0, control console and 3D finite domains viewer only
% 11 January 2002: Version 1, 2D and 3D search tree viewers
% 23 August 2002: Version 2.1.1 recomputations, robust time stamp mechanism for programs
% with arbitrary cuts and exceptions.
% 27 September 2002: Version 2.1.2, 2D finite domains viewer,
% communication of domains with GNU-Prolog syntax
% 14 October 2002: Version 2.1.3 port to Sicstus Prolog
% 24 August 2006: explicit classpath in the call to Java
% 11 October 2006: find_path for infering the classpath

% CHOOSE YOUR VERSION: GNU-Prolog or Sicstus Prolog
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%:- include(gprolog).
:- include(sicstus).

% Toplevel
%%%%%%%%%%
:- initialization(gui).

gui:-
    (argument_value(1,A) -> number_atom(P,A) ; P=27311),
    gui(P).

% gui(Port) or gui(user)
gui(P):-
    write('CLPGUI v2.1.6 by INRIA, Franois Fages (C) 2001, 2002, 2006. On port='), write(P),nl,
    (argument_value(2,nojava)
    -> true
    ;
     number_atom(P,A),
     (find_path(Path) % CLPGUI.classes are in the same directory as clpgui.pl and clpgui(.exe)
      ->
         (environ('!C:',_) % are we under Cygwin ?
          ->
         atom_concat('java -classpath `cygpath -w ',Path,B1),
         atom_concat(B1,'/..` CLPGUI ',B2)
	 ;
	 atom_concat('java -classpath ',Path,B1),
         atom_concat(B1,'/.. CLPGUI ',B2)
         ),
	 atom_concat(B2,A,B)
      ;
        write('Warning: you may have to edit the file clpgui.pl to put the path of CLPGUI.class in the classpath'),
        nl,
	atom_concat('java CLPGUI ',A,B)
      ),
     write(B),nl,
     exec(B)), 
    gui_create_server(P),
    gui_interact0.

find_path(Path) :- 
    (predicate_pathname(gui/0,Path),file_exists(Path),!) % the path of clpgui.pl
    ;
    environ('_',Path). % the path of clpgui(.exe)
    

% Interaction engine
%%%%%%%%%%%%%%%%%%%%

gui_interact0:-
	g_assign(nbvar,0),
	g_assign(names,[]),
	g_assign(variables,[]),
	g_assign(recBacktrack,0),
	g_assign(time,1),
	gui_show_domains(sz),
	catch(gui_interact,X,gui_clear(X)),
	fail.
gui_interact0:- gui_clear(backtrack).

gui_interact:-
        gui_timestamp(s),
	gui_read(N),
	(N==b
	->
	    !, fail
	;
	    (N==a
	    ->
		throw(backtrackInteraction)
	    ;
		(N==c
		->
		    gui_print(continue) 
		;
		    catch((gui_execute(N),gui_interact), backtrackInteraction, fail)))).
gui_interact:- gui_show_domains, fail. 
gui_interact:- gui_interact.

gui_execute(V):- generic_var(V), !, write('ERROR: executing '),write(V),nl,fail.
gui_execute(e):- !, throw(clear).
gui_execute(d):- !, gui_disconnect.
gui_execute(N):-
	g_read(recompute,R),
	(R=0 -> true ; g_assign(recompute,1)),
	gui_name_to_goal(N,G),
	gui_timestampIncr(n,N), % N is identical to the name of the button (unlike trace_search)
	gui_show_domains,
	gui_trace_child(G,N).

gui_clear(X):-
	g_assign(recompute,0),
	(X=recompute -> g_assign(recompute,1), g_assign(recBacktrack,0), g_assign(recomputeLast,0)
	; (X=clear -> true
	  ; gui_print(e))),
	%write(catch(X)),nl,
	gui_interact0.

% Tracing the search tree
%%%%%%%%%%%%%%%%%%%%%%%%%

gui_show:- gui_print(w).
gui_show(T):- gui_timestampIncr(o,T).

gui_trace_search(G):-gui_trace_search(G,_).
gui_trace_search(G,O):-
	gui_goal_to_name(G,N),
	(O==show -> gui_timestampIncr(o,N);gui_timestampIncr(n,N)),
        gui_show_domains,
	gui_trace_child(G,N).

gui_trace_call(G):-gui_trace_call(G,_).
gui_trace_call(G,O):-
	gui_goal_to_name(G,N),
	(O==show -> gui_timestampIncr(o,N);gui_timestampIncr(n,N)),
        gui_show_domains,
	gui_catch_error(G,X,(write(catch(X)),nl,fail)).

gui_trace_child(G):-
	gui_goal_to_name(G,N),
	gui_trace_child(G,N).
gui_trace_child(G,N):-
	gui_catch_error(G,X,(write(catch(X)),nl,fail)),
	gui_goal_to_name(G,R),
	N=..[_|L],
	R=..[_|LR],
	gui_timestampIncr(c,L=LR),
        gui_show_domains.

gui_trace_search_predicate(P/A):-gui_trace_search_predicate(P/A,_).
gui_trace_search_predicate(P/A,O):-
	functor(G,P,A),
	asserta((G:- g_read(gui_hack,H),
		 (G==H
		 ->
		     fail
		 ;
		     !,
		     g_link(gui_hack,G),
		     gui_trace_search(G,O)))).

gui_trace_call_predicate(P/A):-gui_trace_call_predicate(P/A,_).
gui_trace_call_predicate(P/A,O):-
	functor(G,P,A),
	asserta((G:- g_read(gui_hack,H),
		 (G==H
		 ->
		     fail
		 ;
		     !,
		     g_link(gui_hack,G),
		     gui_trace_call(G,O)))).

gui_trace_child_predicate(P/A):-
	functor(G,P,A),
	asserta((G:- g_read(gui_hack,H),
		 (G==H
		 ->
		     fail
		 ;
		     !,
		     g_link(gui_hack,G),
		     gui_trace_child(G)))).

gui_trace_labeling([]).
gui_trace_labeling([X|L]):-
	gui_trace_search(fd_labeling(X)),
	gui_trace_labeling(L).

% timestamp(G) writes on the socket the current depth in the traced search tree, and writes G
% timestampIncr(G) increments the depth
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gui_timestamp(G):-
	g_read(time,T),
	gui_print(t,T,G).
gui_timestamp(G,H):- 
	g_read(time,T),
	gui_print(t,T,G,H).


gui_timestampIncr(G):-
	g_read(recompute,B),	% B=0 if not in recomputation mode,
				% B=-1 if in recomputation mode ignoring which branch to follow
				% B>0 if in recomputation mode following the next Bth branch
	(B=0
	->
	    gui_printTimestampIncr(G)
	;
	    (B=(-1)
	    ->
		gui_read(N),
		((integer(N),N>0)
		->
		    true
		;
		    write('ERROR in recomputation: branch number '),write(N),nl,fail)
	    ;
		N=B),
	    (N=1
	    ->
		(g_read(recomputeLast,0)
		->
		    g_assign(recompute,-1),
		    gui_printTimestampIncr(G)
		;
		    (g_read(recomputeLast,2)
		    ->
			g_assign(recompute,0),
			g_assign(recomputeLast,0)
		    ;
			g_assign(recomputeLast,0)),
		    gui_printTimestampIncr(G),
		    gui_printTimestampIncr(r), 
		    % checks backtracking through a recomputed node
		    (gui_interact; (g_read(recBacktrack,1) -> fail; throw(backtrackInteraction))))
	    ;
		N1 is N-1,
		g_assign(recompute,N1),
		fail)).

gui_timestampIncr(G,H):-
	g_read(recompute,B),	% B=0 if not in recomputation mode,
				% B=-1 if in recomputation mode ignoring which branch to follow
				% B>0 if in recomputation mode following the next Bth branch
	(B=0
	->
	    gui_printTimestampIncr(G), gui_print(H)
	;
	    (B=(-1)
	    ->
		gui_read(N),
		((integer(N),N>0)
		->
		    true
		;
		    write('ERROR in recomputation: branch number '),write(N),nl,fail)
	    ;
		N=B),
	    (N=1
	    ->
		(g_read(recomputeLast,0)
		->
		    g_assign(recompute,-1),
		    gui_printTimestampIncr(G), gui_print(H)
		;
		    (g_read(recomputeLast,2)
		    ->
			g_assign(recompute,0),
			g_assign(recomputeLast,0)
		    ;
			g_assign(recomputeLast,0)),
		    gui_printTimestampIncr(G), gui_printnf(H),
		    gui_printTimestampIncr(r), 
		    % checks backtracking through a recomputed node
		    (gui_interact; (g_read(recBacktrack,1) -> fail; throw(backtrackInteraction))))
	    ;
		N1 is N-1,
		g_assign(recompute,N1),
		fail)).

gui_printTimestampIncr(G):-
	g_read(time,T),
	T1 is T+1,
	g_link(time,T1),
	gui_print(t,T,G).


% Server connection
%%%%%%%%%%%%%%%%%%%

gui_create_server(user):- !,
	add_stream_alias(user_input, input_stream),
	add_stream_alias(user_output, output_stream).
gui_create_server(Port):-
	gui_create_server(localhost,Port).

gui_create_server(Host,Port):-
	catch((socket('AF_INET',NbSocket),
	       socket_bind(NbSocket,'AF_INET'(Host,Port)),
               socket_listen(NbSocket,5),
               socket_accept(NbSocket,_,I,O),
	       add_stream_alias(I, input_stream),
	       add_stream_alias(O, output_stream)),
	      X,write(X)).
gui_create_server(_,_):- gui_disconnect.

gui_disconnect:-
	catch(close(input_stream),_,true),
	catch(close(output_stream),_,true),
	abort.
gui_disconnect:- abort.

% Socket communication
%%%%%%%%%%%%%%%%%%%%%%

gui_read(N):-
	catch((read_term(input_stream,R,[variable_names(VNL)]),
	       name_query_vars(VNL,_)), X, write(catch(X))),
	%write('READ '), write(R), nl, flush_output,
	((R==sv;R==sz;R==sl;R==sn)
	->
	    g_assign(showOption, R), gui_read(N)
	;
	    (R==k
	    ->
		g_assign(recBacktrack,1), gui_read(N)
	    ;
		(functor(R,i,1)
		->
		    R=i(V), gui_name_to_goal(V,W), gui_print(i,V,W), gui_read(N)
		;
		    g_read(recompute,B),
		    (B=0
		    ->
			(R==j
			->
			    throw(recompute)
			;
			    N=R)
		    ; % recomputation mode
			(R==l
			->
			    g_assign(recomputeLast,2),
			    gui_read(N)
			;
			    (R==m
			    ->
				g_assign(recomputeLast,1),
				gui_read(N)
			    ;
				N=R)))))).

%gui_printnf(A):-write(A),nl,fail.
gui_printnf(A):-
	print(output_stream,A), nl(output_stream).

%gui_print(A):-write('PRINT '),write(A),nl,fail.
gui_print(A):-
	print(output_stream,A),nl(output_stream),flush_output(output_stream).

%gui_print(A,B):-write('PRINT '),write(A),write(B),nl,fail.
gui_print(A,B):- 
    print(output_stream,A),nl(output_stream),
    print(output_stream,B),nl(output_stream),flush_output(output_stream).

%gui_print(A,B,C):-write('PRINT '),write(A),write(B),write(C),nl,fail.
gui_print(A,B,C):- 
	print(output_stream,A),nl(output_stream),print(output_stream,B),nl(output_stream),print(output_stream,C),nl(output_stream),flush_output(output_stream).

%gui_print(A,B,C,D):-write('PRINT '),write(A),write(B),write(C),write(D),nl,fail.
gui_print(A,B,C,D):- 
	print(output_stream,A),nl(output_stream),print(output_stream,B),nl(output_stream),print(output_stream,C),nl(output_stream),print(output_stream,D),nl(output_stream),flush_output(output_stream).

gui_print_list([]).
gui_print_list([X|L]):-
	print(output_stream,X),nl(output_stream),
	gui_print_list(L).

gui_print_rev_list([]).
gui_print_rev_list([X|L]):-
	gui_print_rev_list(L),
	print(output_stream,X),nl(output_stream).

gui_sflush:-
	flush_output(output_stream).

% Predicates for associating names to variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gui_varnames:-
	g_read(names,LN),
	g_read(nbvar,N), 
	gui_printnf(a),gui_printnf(N), 
%        gui_print_rev_list(LN),
        gui_print_list(LN),
        gui_sflush.

gui_varnames([]):-gui_varnames.
gui_varnames([V|L]):-
	gui_varnaming(V,_),
	gui_varnames(L).

gui_varnames(LV,LN):-
	gui_addvarnames(LV,LN),
	gui_varnames.

gui_updatevarnames(LV,LN):-
	g_link(variables,LV),
	g_link(names,LN),
	length(LV,N),
	g_link(nbvar,N).
	
gui_addvarnames(V,N):-
        generic_var(V),
        !,
%	g_read(variables,LV),
%       g_read(names,LN),
%       gui_updatevarnames([V|LV],[N|LN]).
        gui_addvarnames([V],[N]).
gui_addvarnames(LV,LN):-
	g_read(variables,L1),
	(L1=[_|_]
	->
	    g_read(names,L2),
%	    append(LV,L1,R1),
%	    append(LN,L2,R2), 
	    append(L1,LV,R1),
	    append(L2,LN,R2), 
	    gui_updatevarnames(R1,R2)
	;
	    gui_updatevarnames(LV,LN)).

gui_varnaming(V,N):-
	g_read(variable_number,I),
	number_atom(I,AI),
	atom_concat('V',AI,N),
	I1 is I+1,
	g_link(variable_number,I1),
	gui_addvarnames([V],[N]).

gui_varname(V,N):- 
	g_read(names,LN), 
	g_read(variables,LV),
	gui_nth(LV,V,LN,N).

gui_nth([A|_],B,[C|_],D):- (A==B;C==D),!,A=B,C=D.
gui_nth([_|L],B,[_|K],D):- gui_nth(L,B,K,D).

% Goal naming
%%%%%%%%%%%%%

gui_button(G):-
	gui_goal_to_name(G,N),
	gui_print(b,N).

gui_bagof_buttons(Template,Goal):-
	bagof(Template,Goal,L),
	gui_button_list(L).
gui_button_list([]).
gui_button_list([G|L]):- call(gui_button(G)), gui_button_list(L).

gui_goal_to_name(V,N):-
    g_assign(newvar,0),
    gui_goal_to_name2(V,N),
    (g_read(newvar,0) -> true ; gui_varnames).

gui_goal_to_name2(V,N):- generic_var(V), !,
	(gui_varname(V,N) -> true; gui_varnaming(V,N),g_assign(newvar,1)).
gui_goal_to_name2([],[]):- !.
gui_goal_to_name2([T|L],[N|K]):- !,
	gui_goal_to_name2(T,N),
	gui_goal_to_name2(L,K).
gui_goal_to_name2(T,N):-
	T=..[F|L],
	gui_goal_to_name2(L,K),
	N=..[F|K].

gui_name_to_goal(N,_):- generic_var(N), !. % read a wildcard
gui_name_to_goal('$VARNAME'(N),V):- !,
	g_read(names,LN),
	g_read(variables,LV),
	(gui_nth(LV,V,LN,N)
	-> true
	;
	    gui_updatevarnames([V|LV],[N|LN]),
	    gui_varnames). 
gui_name_to_goal(N,V):- gui_varname(V,N), !.
gui_name_to_goal([],[]):- !.
gui_name_to_goal([N|K],[T|L]):- !,
	gui_name_to_goal(N,T),
	gui_name_to_goal(K,L).
gui_name_to_goal(N,T):-
	N=..[F|K],
	gui_name_to_goal(K,L),
	T=..[F|L].

% Visualization of variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gui_no_trace_backtracking:- gui_print(nt).
gui_trace_backtracking:- gui_print(tr).

gui_show_no_domain:- g_assign(showOption, sn), gui_print(sn).
gui_show_values:- g_assign(showOption, sv), gui_print(sv).
gui_show_sizes:- g_assign(showOption, sz), gui_print(sz).
gui_show_intervals:- g_assign(showOption, sl), gui_print(sl).

gui_show_domains(R):-
	g_assign(showOption, R), 
	((R==sv;R==sz;R==sl;R==sn)
	->
	    gui_print(R)
	;
	    write('WARNING unknown option: show_domains('), write(R), write(')'), nl,
	    g_assign(showOption, sz), gui_print(sz)).

gui_show_domains:-
	g_read(showOption,O),
	((O==sn)
	->
	    true
	;
	    g_read(variables,L),
	    gui_show_domains(O, L),
	    gui_sflush).

gui_show_domains(sv, L):- !, g_read(nbvar,N), gui_printnf(v), gui_printnf(N), gui_showDomains(L).
gui_show_domains(sl, L):- !, g_read(nbvar,N), gui_printnf(l), gui_printnf(N), gui_showIntervals(L).
gui_show_domains(sn, _):- !.
gui_show_domains(_, L):- g_read(nbvar,N), gui_printnf(z), gui_printnf(N), gui_showSizes(L).

gui_showSizes([]).
gui_showSizes([X|L]):-
    (gui_fd(X) -> fd_size(X,S) ; S=0),
    gui_printnf(S),
    gui_showSizes(L).

gui_showDomains([]).
gui_showDomains([X|L]):-
    (gui_fd(X)
    ->
	gui_print_dom(X)
    ;
	gui_printnf(0)),
    gui_showDomains(L).

gui_showIntervals([]).
gui_showIntervals([X|L]):-
    (gui_fd(X)
    ->
	fd_size(X,S),
	fd_min(X,M),
	fd_max(X,N),
	gui_print(S,M,N)
    ;
	gui_printnf(0)),
    gui_showIntervals(L).


